@@ -36,16 +36,12 @@
  * I don't think it's really that important though.
  *
  *
- * Please see the file LICENSE in the source's root directory.
+ * Please see the file LICENSE.txt in the source's root directory.
  *
  * This file written by Travis Wells, based on the GRP archiver by
  *  Ryan C. Gordon.
  */
 
-#if HAVE_CONFIG_H
-#  include <config.h>
-#endif
-
 #if (defined PHYSFS_SUPPORTS_WAD)
 
 #include <stdio.h>
@@ -80,86 +76,19 @@ typedef struct
 } WADfileinfo;
 
 
-static void WAD_dirClose(DirHandle *h);
-static PHYSFS_sint64 WAD_read(FileHandle *handle, void *buffer,
-                              PHYSFS_uint32 objSize, PHYSFS_uint32 objCount);
-static PHYSFS_sint64 WAD_write(FileHandle *handle, const void *buffer,
-                               PHYSFS_uint32 objSize, PHYSFS_uint32 objCount);
-static int WAD_eof(FileHandle *handle);
-static PHYSFS_sint64 WAD_tell(FileHandle *handle);
-static int WAD_seek(FileHandle *handle, PHYSFS_uint64 offset);
-static PHYSFS_sint64 WAD_fileLength(FileHandle *handle);
-static int WAD_fileClose(FileHandle *handle);
-static int WAD_isArchive(const char *filename, int forWriting);
-static DirHandle *WAD_openArchive(const char *name, int forWriting);
-static LinkedStringList *WAD_enumerateFiles(DirHandle *h,
-                                            const char *dirname,
-                                            int omitSymLinks);
-static int WAD_exists(DirHandle *h, const char *name);
-static int WAD_isDirectory(DirHandle *h, const char *name, int *fileExists);
-static int WAD_isSymLink(DirHandle *h, const char *name, int *fileExists);
-static PHYSFS_sint64 WAD_getLastModTime(DirHandle *h, const char *n, int *e);
-static FileHandle *WAD_openRead(DirHandle *h, const char *name, int *exist);
-static FileHandle *WAD_openWrite(DirHandle *h, const char *name);
-static FileHandle *WAD_openAppend(DirHandle *h, const char *name);
-static int WAD_remove(DirHandle *h, const char *name);
-static int WAD_mkdir(DirHandle *h, const char *name);
-
-const PHYSFS_ArchiveInfo __PHYSFS_ArchiveInfo_WAD =
+static void WAD_dirClose(dvoid *opaque)
 {
-    "WAD",
-    WAD_ARCHIVE_DESCRIPTION,
-    "Travis Wells <traviswells@mchsi.com>",
-    "http://www.3dmm2.com/doom/",
-};
-
-
-static const FileFunctions __PHYSFS_FileFunctions_WAD =
-{
-    WAD_read,       /* read() method       */
-    WAD_write,      /* write() method      */
-    WAD_eof,        /* eof() method        */
-    WAD_tell,       /* tell() method       */
-    WAD_seek,       /* seek() method       */
-    WAD_fileLength, /* fileLength() method */
-    WAD_fileClose   /* fileClose() method  */
-};
-
-
-const DirFunctions __PHYSFS_DirFunctions_WAD =
-{
-    &__PHYSFS_ArchiveInfo_WAD,
-    WAD_isArchive,          /* isArchive() method      */
-    WAD_openArchive,        /* openArchive() method    */
-    WAD_enumerateFiles,     /* enumerateFiles() method */
-    WAD_exists,             /* exists() method         */
-    WAD_isDirectory,        /* isDirectory() method    */
-    WAD_isSymLink,          /* isSymLink() method      */
-    WAD_getLastModTime,     /* getLastModTime() method */
-    WAD_openRead,           /* openRead() method       */
-    WAD_openWrite,          /* openWrite() method      */
-    WAD_openAppend,         /* openAppend() method     */
-    WAD_remove,             /* remove() method         */
-    WAD_mkdir,              /* mkdir() method          */
-    WAD_dirClose            /* dirClose() method       */
-};
-
-
-
-static void WAD_dirClose(DirHandle *h)
-{
-    WADinfo *info = ((WADinfo *) h->opaque);
-    free(info->filename);
-    free(info->entries);
-    free(info);
-    free(h);
+    WADinfo *info = ((WADinfo *) opaque);
+    allocator.Free(info->filename);
+    allocator.Free(info->entries);
+    allocator.Free(info);
 } /* WAD_dirClose */
 
 
-static PHYSFS_sint64 WAD_read(FileHandle *handle, void *buffer,
+static PHYSFS_sint64 WAD_read(fvoid *opaque, void *buffer,
                               PHYSFS_uint32 objSize, PHYSFS_uint32 objCount)
 {
-    WADfileinfo *finfo = (WADfileinfo *) (handle->opaque);
+    WADfileinfo *finfo = (WADfileinfo *) opaque;
     WADentry *entry = finfo->entry;
     PHYSFS_uint32 bytesLeft = entry->size - finfo->curPos;
     PHYSFS_uint32 objsLeft = (bytesLeft / objSize);
@@ -176,30 +105,30 @@ static PHYSFS_sint64 WAD_read(FileHandle *handle, void *buffer,
 } /* WAD_read */
 
 
-static PHYSFS_sint64 WAD_write(FileHandle *handle, const void *buffer,
+static PHYSFS_sint64 WAD_write(fvoid *opaque, const void *buffer,
                                PHYSFS_uint32 objSize, PHYSFS_uint32 objCount)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, -1);
 } /* WAD_write */
 
 
-static int WAD_eof(FileHandle *handle)
+static int WAD_eof(fvoid *opaque)
 {
-    WADfileinfo *finfo = (WADfileinfo *) (handle->opaque);
+    WADfileinfo *finfo = (WADfileinfo *) opaque;
     WADentry *entry = finfo->entry;
     return(finfo->curPos >= entry->size);
 } /* WAD_eof */
 
 
-static PHYSFS_sint64 WAD_tell(FileHandle *handle)
+static PHYSFS_sint64 WAD_tell(fvoid *opaque)
 {
-    return(((WADfileinfo *) (handle->opaque))->curPos);
+    return(((WADfileinfo *) opaque)->curPos);
 } /* WAD_tell */
 
 
-static int WAD_seek(FileHandle *handle, PHYSFS_uint64 offset)
+static int WAD_seek(fvoid *opaque, PHYSFS_uint64 offset)
 {
-    WADfileinfo *finfo = (WADfileinfo *) (handle->opaque);
+    WADfileinfo *finfo = (WADfileinfo *) opaque;
     WADentry *entry = finfo->entry;
     int rc;
 
@@ -213,19 +142,18 @@ static int WAD_seek(FileHandle *handle, PHYSFS_uint64 offset)
 } /* WAD_seek */
 
 
-static PHYSFS_sint64 WAD_fileLength(FileHandle *handle)
+static PHYSFS_sint64 WAD_fileLength(fvoid *opaque)
 {
-    WADfileinfo *finfo = ((WADfileinfo *) handle->opaque);
+    WADfileinfo *finfo = (WADfileinfo *) opaque;
     return((PHYSFS_sint64) finfo->entry->size);
 } /* WAD_fileLength */
 
 
-static int WAD_fileClose(FileHandle *handle)
+static int WAD_fileClose(fvoid *opaque)
 {
-    WADfileinfo *finfo = ((WADfileinfo *) handle->opaque);
+    WADfileinfo *finfo = (WADfileinfo *) opaque;
     BAIL_IF_MACRO(!__PHYSFS_platformClose(finfo->handle), NULL, 0);
-    free(finfo);
-    free(handle);
+    allocator.Free(finfo);
     return(1);
 } /* WAD_fileClose */
 
@@ -287,27 +215,19 @@ static int WAD_isArchive(const char *filename, int forWriting)
 
 static int wad_entry_cmp(void *_a, PHYSFS_uint32 one, PHYSFS_uint32 two)
 {
-    if (one != two)
-    {
-        const WADentry *a = (const WADentry *) _a;
-        return(strcmp(a[one].name, a[two].name));
-    } /* if */
-
-    return 0;
+    WADentry *a = (WADentry *) _a;
+    return(strcmp(a[one].name, a[two].name));
 } /* wad_entry_cmp */
 
 
 static void wad_entry_swap(void *_a, PHYSFS_uint32 one, PHYSFS_uint32 two)
 {
-    if (one != two)
-    {
-        WADentry tmp;
-        WADentry *first = &(((WADentry *) _a)[one]);
-        WADentry *second = &(((WADentry *) _a)[two]);
-        memcpy(&tmp, first, sizeof (WADentry));
-        memcpy(first, second, sizeof (WADentry));
-        memcpy(second, &tmp, sizeof (WADentry));
-    } /* if */
+    WADentry tmp;
+    WADentry *first = &(((WADentry *) _a)[one]);
+    WADentry *second = &(((WADentry *) _a)[two]);
+    memcpy(&tmp, first, sizeof (WADentry));
+    memcpy(first, second, sizeof (WADentry));
+    memcpy(second, &tmp, sizeof (WADentry));
 } /* wad_entry_swap */
 
 
@@ -323,7 +243,7 @@ static int wad_load_entries(const char *name, int forWriting, WADinfo *info)
 
     BAIL_IF_MACRO(!wad_open(name, forWriting, &fh, &fileCount,&directoryOffset), NULL, 0);
     info->entryCount = fileCount;
-    info->entries = (WADentry *) malloc(sizeof (WADentry) * fileCount);
+    info->entries = (WADentry *) allocator.Malloc(sizeof(WADentry)*fileCount);
     if (info->entries == NULL)
     {
         __PHYSFS_platformClose(fh);
@@ -365,94 +285,71 @@ static int wad_load_entries(const char *name, int forWriting, WADinfo *info)
 } /* wad_load_entries */
 
 
-static DirHandle *WAD_openArchive(const char *name, int forWriting)
+static void *WAD_openArchive(const char *name, int forWriting)
 {
-    WADinfo *info;
-    DirHandle *retval = malloc(sizeof (DirHandle));
     PHYSFS_sint64 modtime = __PHYSFS_platformGetLastModTime(name);
+    WADinfo *info = (WADinfo *) allocator.Malloc(sizeof (WADinfo));
 
-    BAIL_IF_MACRO(retval == NULL, ERR_OUT_OF_MEMORY, NULL);
-    info = retval->opaque = malloc(sizeof (WADinfo));
-    if (info == NULL)
-    {
-        __PHYSFS_setError(ERR_OUT_OF_MEMORY);
-        goto WAD_openArchive_failed;
-    } /* if */
-
+    BAIL_IF_MACRO(info == NULL, ERR_OUT_OF_MEMORY, NULL);
     memset(info, '\0', sizeof (WADinfo));
 
-    info->filename = (char *) malloc(strlen(name) + 1);
-    if (info->filename == NULL)
-    {
-        __PHYSFS_setError(ERR_OUT_OF_MEMORY);
-        goto WAD_openArchive_failed;
-    } /* if */
+    info->filename = (char *) allocator.Malloc(strlen(name) + 1);
+    GOTO_IF_MACRO(!info->filename, ERR_OUT_OF_MEMORY, WAD_openArchive_failed);
 
     if (!wad_load_entries(name, forWriting, info))
         goto WAD_openArchive_failed;
 
     strcpy(info->filename, name);
     info->last_mod_time = modtime;
-    retval->funcs = &__PHYSFS_DirFunctions_WAD;
-    return(retval);
+    return(info);
 
 WAD_openArchive_failed:
-    if (retval != NULL)
+    if (info != NULL)
     {
-        if (retval->opaque != NULL)
-        {
-            if (info->filename != NULL)
-                free(info->filename);
-            if (info->entries != NULL)
-                free(info->entries);
-            free(info);
-        } /* if */
-        free(retval);
+        if (info->filename != NULL)
+            allocator.Free(info->filename);
+        if (info->entries != NULL)
+            allocator.Free(info->entries);
+        allocator.Free(info);
     } /* if */
 
     return(NULL);
 } /* WAD_openArchive */
 
 
-static LinkedStringList *WAD_enumerateFiles(DirHandle *h,
-                                            const char *dirname,
-                                            int omitSymLinks)
+static void WAD_enumerateFiles(dvoid *opaque, const char *dname,
+                               int omitSymLinks, PHYSFS_EnumFilesCallback cb,
+                               const char *origdir, void *callbackdata)
 {
-    WADinfo *info = ((WADinfo *) h->opaque);
+    WADinfo *info = ((WADinfo *) opaque);
     WADentry *entry = info->entries;
-    LinkedStringList *retval = NULL, *p = NULL;
     PHYSFS_uint32 max = info->entryCount;
     PHYSFS_uint32 i;
+    const char *name;
     char *sep;
 
-    if (dirname[0] == 0)
+    if (*dname == '\0')  /* root directory enumeration? */
     {
         for (i = 0; i < max; i++, entry++)
         {
-            if (strchr(entry->name, '/') == NULL)
-            {
-                retval = __PHYSFS_addToLinkedStringList(retval, &p,
-                                                        entry->name, -1);
-            } /* if */
+            name = entry->name;
+            if (strchr(name, '/') == NULL)
+                cb(callbackdata, origdir, name);
         } /* for */
     } /* if */
     else
     {
         for (i = 0; i < max; i++, entry++)
         {
-            sep = strchr(entry->name, '/');
+            name = entry->name;
+            sep = strchr(name, '/');
             if (sep != NULL)
             {
-                if (strncmp(dirname, entry->name, (sep-entry->name)) == 0)
-                {
-                    retval = __PHYSFS_addToLinkedStringList(retval, &p,
-                                                            sep + 1, -1);
-                } /* if */
+                if (strncmp(dname, name, (sep - name)) == 0)
+                    cb(callbackdata, origdir, sep + 1);
             } /* if */
         } /* for */
     } /* else */
-    
-    return(retval);
 } /* WAD_enumerateFiles */
 
 
@@ -480,15 +377,15 @@ static WADentry *wad_find_entry(WADinfo *info, const char *name)
 } /* wad_find_entry */
 
 
-static int WAD_exists(DirHandle *h, const char *name)
+static int WAD_exists(dvoid *opaque, const char *name)
 {
-    return(wad_find_entry(((WADinfo *) h->opaque), name) != NULL);
+    return(wad_find_entry(((WADinfo *) opaque), name) != NULL);
 } /* WAD_exists */
 
 
-static int WAD_isDirectory(DirHandle *h, const char *name, int *fileExists)
+static int WAD_isDirectory(dvoid *opaque, const char *name, int *fileExists)
 {
-    WADentry *entry = wad_find_entry(((WADinfo *) h->opaque), name);
+    WADentry *entry = wad_find_entry(((WADinfo *) opaque), name);
     if (entry != NULL)
     {
         char *n;
@@ -516,32 +413,31 @@ static int WAD_isDirectory(DirHandle *h, const char *name, int *fileExists)
 } /* WAD_isDirectory */
 
 
-static int WAD_isSymLink(DirHandle *h, const char *name, int *fileExists)
+static int WAD_isSymLink(dvoid *opaque, const char *name, int *fileExists)
 {
-    *fileExists = WAD_exists(h, name);
+    *fileExists = WAD_exists(opaque, name);
     return(0);  /* never symlinks in a wad. */
 } /* WAD_isSymLink */
 
 
-static PHYSFS_sint64 WAD_getLastModTime(DirHandle *h,
+static PHYSFS_sint64 WAD_getLastModTime(dvoid *opaque,
                                         const char *name,
                                         int *fileExists)
 {
-    WADinfo *info = ((WADinfo *) h->opaque);
+    WADinfo *info = ((WADinfo *) opaque);
     PHYSFS_sint64 retval = -1;
 
     *fileExists = (wad_find_entry(info, name) != NULL);
     if (*fileExists)  /* use time of WAD itself in the physical filesystem. */
-        retval = ((WADinfo *) h->opaque)->last_mod_time;
+        retval = info->last_mod_time;
 
     return(retval);
 } /* WAD_getLastModTime */
 
 
-static FileHandle *WAD_openRead(DirHandle *h, const char *fnm, int *fileExists)
+static fvoid *WAD_openRead(dvoid *opaque, const char *fnm, int *fileExists)
 {
-    WADinfo *info = ((WADinfo *) h->opaque);
-    FileHandle *retval;
+    WADinfo *info = ((WADinfo *) opaque);
     WADfileinfo *finfo;
     WADentry *entry;
 
@@ -549,56 +445,81 @@ static FileHandle *WAD_openRead(DirHandle *h, const char *fnm, int *fileExists)
     *fileExists = (entry != NULL);
     BAIL_IF_MACRO(entry == NULL, NULL, NULL);
 
-    retval = (FileHandle *) malloc(sizeof (FileHandle));
-    BAIL_IF_MACRO(retval == NULL, ERR_OUT_OF_MEMORY, NULL);
-    finfo = (WADfileinfo *) malloc(sizeof (WADfileinfo));
-    if (finfo == NULL)
-    {
-        free(retval);
-        BAIL_MACRO(ERR_OUT_OF_MEMORY, NULL);
-    } /* if */
+    finfo = (WADfileinfo *) allocator.Malloc(sizeof (WADfileinfo));
+    BAIL_IF_MACRO(finfo == NULL, ERR_OUT_OF_MEMORY, NULL);
 
     finfo->handle = __PHYSFS_platformOpenRead(info->filename);
     if ( (finfo->handle == NULL) ||
          (!__PHYSFS_platformSeek(finfo->handle, entry->startPos)) )
     {
-        free(finfo);
-        free(retval);
+        allocator.Free(finfo);
         return(NULL);
     } /* if */
 
     finfo->curPos = 0;
     finfo->entry = entry;
-    retval->opaque = (void *) finfo;
-    retval->funcs = &__PHYSFS_FileFunctions_WAD;
-    retval->dirHandle = h;
-    return(retval);
+    return(finfo);
 } /* WAD_openRead */
 
 
-static FileHandle *WAD_openWrite(DirHandle *h, const char *name)
+static fvoid *WAD_openWrite(dvoid *opaque, const char *name)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, NULL);
 } /* WAD_openWrite */
 
 
-static FileHandle *WAD_openAppend(DirHandle *h, const char *name)
+static fvoid *WAD_openAppend(dvoid *opaque, const char *name)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, NULL);
 } /* WAD_openAppend */
 
 
-static int WAD_remove(DirHandle *h, const char *name)
+static int WAD_remove(dvoid *opaque, const char *name)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, 0);
 } /* WAD_remove */
 
 
-static int WAD_mkdir(DirHandle *h, const char *name)
+static int WAD_mkdir(dvoid *opaque, const char *name)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, 0);
 } /* WAD_mkdir */
 
+
+const PHYSFS_ArchiveInfo __PHYSFS_ArchiveInfo_WAD =
+{
+    "WAD",
+    WAD_ARCHIVE_DESCRIPTION,
+    "Travis Wells <traviswells@mchsi.com>",
+    "http://www.3dmm2.com/doom/",
+};
+
+
+const PHYSFS_Archiver __PHYSFS_Archiver_WAD =
+{
+    &__PHYSFS_ArchiveInfo_WAD,
+    WAD_isArchive,          /* isArchive() method      */
+    WAD_openArchive,        /* openArchive() method    */
+    WAD_enumerateFiles,     /* enumerateFiles() method */
+    WAD_exists,             /* exists() method         */
+    WAD_isDirectory,        /* isDirectory() method    */
+    WAD_isSymLink,          /* isSymLink() method      */
+    WAD_getLastModTime,     /* getLastModTime() method */
+    WAD_openRead,           /* openRead() method       */
+    WAD_openWrite,          /* openWrite() method      */
+    WAD_openAppend,         /* openAppend() method     */
+    WAD_remove,             /* remove() method         */
+    WAD_mkdir,              /* mkdir() method          */
+    WAD_dirClose,           /* dirClose() method       */
+    WAD_read,               /* read() method           */
+    WAD_write,              /* write() method          */
+    WAD_eof,                /* eof() method            */
+    WAD_tell,               /* tell() method           */
+    WAD_seek,               /* seek() method           */
+    WAD_fileLength,         /* fileLength() method     */
+    WAD_fileClose           /* fileClose() method      */
+};
+
 #endif  /* defined PHYSFS_SUPPORTS_WAD */
 
 /* end of wad.c ... */
